#include <allegro.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#define COLOR_RED 0
#define COLOR_YEL 1
#define COLOR_GRN 2
#define COLOR_BLU 3

#define DIR_UP 0
#define DIR_UR 1
#define DIR_DR 2
#define DIR_DN 3
#define DIR_DL 4
#define DIR_UL 5

typedef struct {
	int outer;
	int inner;
	int tension[6];
} PIECE;

BITMAP* make_piece (PIECE p, BITMAP *pieces, bool large) {
    if (large) {
		BITMAP *b = create_bitmap(45,44);
		clear_bitmap(b);
		blit(pieces,b,1,58,0,0,45,44);

		if (p.outer == COLOR_RED) masked_blit(pieces,b,1,103,0,0,45,44);
		if (p.outer == COLOR_YEL) masked_blit(pieces,b,47,103,0,0,45,44);
		if (p.outer == COLOR_GRN) masked_blit(pieces,b,93,103,0,0,45,44);

		if (p.inner == COLOR_RED) masked_blit(pieces,b,1,148,0,0,45,44);
		if (p.inner == COLOR_YEL) masked_blit(pieces,b,47,148,0,0,45,44);
		if (p.inner == COLOR_GRN) masked_blit(pieces,b,93,148,0,0,45,44);

		if (p.tension[DIR_UP] == COLOR_RED) masked_blit(pieces,b,1,193,0,0,45,44);
		if (p.tension[DIR_UP] == COLOR_YEL) masked_blit(pieces,b,47,193,0,0,45,44);
		if (p.tension[DIR_UP] == COLOR_GRN) masked_blit(pieces,b,93,193,0,0,45,44);

		if (p.tension[DIR_UR] == COLOR_RED) masked_blit(pieces,b,1,238,0,0,45,44);
		if (p.tension[DIR_UR] == COLOR_YEL) masked_blit(pieces,b,47,238,0,0,45,44);
		if (p.tension[DIR_UR] == COLOR_GRN) masked_blit(pieces,b,93,238,0,0,45,44);

		if (p.tension[DIR_DR] == COLOR_RED) masked_blit(pieces,b,1,283,0,0,45,44);
		if (p.tension[DIR_DR] == COLOR_YEL) masked_blit(pieces,b,47,283,0,0,45,44);
		if (p.tension[DIR_DR] == COLOR_GRN) masked_blit(pieces,b,93,283,0,0,45,44);

		if (p.tension[DIR_DN] == COLOR_RED) masked_blit(pieces,b,1,328,0,0,45,44);
		if (p.tension[DIR_DN] == COLOR_YEL) masked_blit(pieces,b,47,328,0,0,45,44);
		if (p.tension[DIR_DN] == COLOR_GRN) masked_blit(pieces,b,93,328,0,0,45,44);

		if (p.tension[DIR_DL] == COLOR_RED) masked_blit(pieces,b,1,373,0,0,45,44);
		if (p.tension[DIR_DL] == COLOR_YEL) masked_blit(pieces,b,47,373,0,0,45,44);
		if (p.tension[DIR_DL] == COLOR_GRN) masked_blit(pieces,b,93,373,0,0,45,44);

		if (p.tension[DIR_UL] == COLOR_RED) masked_blit(pieces,b,1,418,0,0,45,44);
		if (p.tension[DIR_UL] == COLOR_YEL) masked_blit(pieces,b,47,418,0,0,45,44);
		if (p.tension[DIR_UL] == COLOR_GRN) masked_blit(pieces,b,93,418,0,0,45,44);
		
		return b;
	} else {
		BITMAP *b = create_bitmap(19,18);
		clear_bitmap(b);
		blit(pieces,b,1,1,0,0,19,18);

		if (p.outer == COLOR_RED) masked_blit(pieces,b,1,20,0,0,19,18);
		if (p.outer == COLOR_YEL) masked_blit(pieces,b,21,20,0,0,19,18);
		if (p.outer == COLOR_GRN) masked_blit(pieces,b,41,20,0,0,19,18);
	
		if (p.inner == COLOR_RED) masked_blit(pieces,b,1,39,0,0,19,18);
		if (p.inner == COLOR_YEL) masked_blit(pieces,b,21,39,0,0,19,18);
		if (p.inner == COLOR_GRN) masked_blit(pieces,b,41,39,0,0,19,18);
		return b;
	}
}

inline void board_to_screen(int h, int k, int &x, int &y) {
	x = 14+(h-2)*32+(k-1)*31;
	y = 212-(h-2)*20+(k-1)*21;
}

inline void get_relative_h_k(int dir, int &h, int &k) {
	switch (dir) {
		case DIR_UP:
			h = 1; k = -1;
			break;
		case DIR_UL:
			h = 0; k = -1;
			break;
		case DIR_DL:
			h = -1; k = 0;
			break;
		case DIR_DN:
			h = -1; k = 1;
			break;
		case DIR_DR:
			h = 0; k = 1;
			break;
		case DIR_UR:
			h = 1; k = 0;
	}
}

inline bool exists(PIECE a) {
	//this function is only needed to check for dne, and since -1 can't show
	//up in any other context, we just need to test if the outers are equal
	if (a.outer == -1) return false;
	else return true;
}

const PIECE dne = {-1,-1,{-1,-1,-1,-1,-1,-1}}; //This piece means no piece.
int dead_x = -1; //Board x coordinate of the gameover cell
int dead_y = -1; //Board y corrdinate of the gameover cell
int score; //Score
PIECE board[9][8]; //The game board.
PIECE queue[4]; //The next pieces queue.
int cursor_x; //Board x coordinate of the cursor
int cursor_y; //Board y coordinate of the cursor

void do_page(BITMAP *pieces, BITMAP *hud, BITMAP *page) { //Updates page
	//first blit hud to page
	blit(hud,page,0,0,0,0,640,480);
	//now we need to blit all the pieces onto the board
	//this is a bit complicated, try to follow
	for (int h = 1; h <= 9; h++) {
		for (int k = 1; k <= 8; k++) {
			// these two cells don't exist
			if ((h == 1 && k == 1) || (h == 9 && k == 8)) continue;
			if (exists(board[h-1][k-1])) {
				int x = 0, y = 0;
				board_to_screen(h,k,x,y);
				BITMAP *p = make_piece(board[h-1][k-1],pieces,true);
				masked_blit(p,page,0,0,x,y,45,44);
				destroy_bitmap(p);
			}
		}
	}
	//blits pieces to the queue
	BITMAP* b = make_piece(queue[0], pieces, true);
	masked_blit (b,page,0,0,491,432,45,44);
	
	b = make_piece(queue[1], pieces, true);
	masked_blit (b,page,0,0,542,432,45,44);
	
	b = make_piece(queue[2], pieces, false);
	masked_blit (b,page,0,0,593,458,19,18);
	
	b = make_piece(queue[3], pieces, false);
	masked_blit (b,page,0,0,617,458,19,18);
	
	//draws the cursor
	int cursor_screen_x;
	int cursor_screen_y;
	if (exists(board[cursor_x-1][cursor_y-1])) {
		board_to_screen(cursor_x,cursor_y,cursor_screen_x,cursor_screen_y);
		masked_blit(pieces,page,139,103,cursor_screen_x,cursor_screen_y,
					45,44);
	}
	else {
		board_to_screen(cursor_x,cursor_y,cursor_screen_x,cursor_screen_y);
		masked_blit(pieces,page,139,148,cursor_screen_x,cursor_screen_y,
					45,44);
	}
	
	//score display
	textprintf_ex(page, font, 0, 0, makecol(255,255,255), -1, "Score: %d", score);
	
	//finally, blit gameover symbol over gameover square
	if (dead_x != -1 && dead_y != -1) {
		int dead_screen_x, dead_screen_y;
		board_to_screen(dead_x,dead_y,dead_screen_x,dead_screen_y);
		masked_blit(pieces,page,139,193,dead_screen_x,dead_screen_y,45,44);
	}
}		

int update_tension(void) {
	int game = 0;
	for (int i = 0; i < 9; i++) {
		for (int j = 0; j < 8; j++) {
			if (!exists(board[i][j])) continue;
			for (int dir = 0; dir < 6; dir++) {
				int h,k;
				get_relative_h_k(dir,h,k);
				if (i+h < 9 && i+h >= 0 && j+k < 8 && j+k >= 0 &&
					exists(board[i+h][j+k])) {
					int diff1 = abs(board[i][j].inner - board[i+h][j+k].inner);
					int diff2 = abs(board[i][j].outer - board[i+h][j+k].outer);
					int diff = diff1+diff2;
					int tension = floor((double)diff / 2);
					board[i][j].tension[dir] = 3 - tension;
				}
				else board[i][j].tension[dir] = COLOR_BLU;
			}
		}
	}
	//now search the board for gameover conditions
	//also calculate tension bonus
	int tension_bonus = 0;
	int board_full = 1;
	for (int i = 0; i < 9; i++) {
		for (int j = 0; j < 8; j++) {
			if ((i == 0 && j == 0) || (i == 8 && j == 7)) continue;
			if (!exists(board[i][j])) {
				board_full = 0;
				continue;
			}
			int n = board[i][j].tension[0] +
					board[i][j].tension[1] +
					board[i][j].tension[2] +
					board[i][j].tension[3] +
					board[i][j].tension[4] +
					board[i][j].tension[5];
			if (n <= 10) {
				dead_x = i+1;
				dead_y = j+1;
				return 1;
			}
			tension_bonus += n;
		}
	}
	if (board_full == 1) {
		game = 1;
		score += tension_bonus;
	}
	return game;
}
						
void init_board(void) {
	for (int i = 0; i < 9; i++) {
		for (int j = 0; j < 8; j++) {
			board[i][j] = dne; //correct behavior
//			if ((j == 0 && i == 0) || (j == 8 && i == 7)) {
//				board[i][j] = dne;
//				continue;
//			}
//			PIECE p = {rand()%5-1, rand()%4, {3,3,3,3,3,3}};
//			board[i][j] = p; 
//			update_tension(); //for testing do_page and update_tension
		}
	}
	for (int i = 0; i < 4; i++) {
		PIECE p = {rand()%4, rand()%4, {3,3,3,3,3,3}};
		queue[i] = p;
	}
	cursor_x = 5; cursor_y = 5;
}

void stabilize_before_exit () {
}

int main (int nArgs, char *pszArgs) {
	install_allegro(SYSTEM_AUTODETECT, &errno, atexit);
	install_keyboard();
	install_mouse();
	atexit(stabilize_before_exit);

	srand( time(NULL) );

	set_color_depth( desktop_color_depth() );
	if ( set_gfx_mode(GFX_AUTODETECT_WINDOWED,640,480,0,0) ) {
    	allegro_message("Could not create the graphics window. Apologies.");
		exit(1);
	}        

	set_window_title ("Hexagonia");
	
	BITMAP *pieces = load_bitmap("hexagons.bmp",NULL);

/*	//Uncomment this code to recompile hextest.exe
	while (!keypressed()) {
		PIECE p = {rand()%4,rand()%4, {rand()%4,rand()%4,rand()%4,rand()%4,rand()%4,rand()%4}};
		BITMAP *b = make_piece(p, pieces, rand()%2==1);
		masked_blit (b,screen,0,0,rand()%595,rand()%436,b->w,b->h);
		destroy_bitmap(b);
		rest(1);
	}


	destroy_bitmap(pieces);


	return 0;
*/ 
	
	init_board();

	BITMAP *page = create_bitmap(640,480);
	clear_bitmap(page);
	
	BITMAP *hud = load_bitmap("hud.bmp",NULL);
	
	int game = 0;
	score = 0;

	do {
		do_page(pieces, hud, page);
		blit(page,screen,0,0,0,0,640,480);
		int h, k;
		while (!key[KEY_ESC] && !game) {
			int key = readkey();
			if (key >> 8 == KEY_8_PAD) {
				get_relative_h_k(DIR_UP,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_9_PAD) {
				get_relative_h_k(DIR_UR,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_3_PAD) {
				get_relative_h_k(DIR_DR,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_2_PAD) {
				get_relative_h_k(DIR_DN,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_1_PAD) {
				get_relative_h_k(DIR_DL,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_7_PAD) {
				get_relative_h_k(DIR_UL,h,k);
				if (cursor_x+h-1 < 9 && cursor_x+h-1 >= 0 && cursor_y+k-1 < 8 &&
					cursor_y+k-1 >= 0 && !(cursor_x+h == 1 && cursor_y+k == 1) &&
					!(cursor_x+h == 9 && cursor_y+k == 8)) {
					cursor_x += h;
					cursor_y += k;
					do_page(pieces, hud, page);
					blit(page,screen,0,0,0,0,640,480);
				}
			}
			if (key >> 8 == KEY_0_PAD && !exists(board[cursor_x-1][cursor_y-1])) {
				board[cursor_x-1][cursor_y-1] = queue[0];
				queue[0]=queue[1];
				queue[1]=queue[2];
				queue[2]=queue[3];
				
				PIECE p = {rand()%4, rand()%4, {3,3,3,3,3,3}};
				queue[3] = p;
				
				game = update_tension();
				
				score++;
				
				do_page(pieces, hud, page);
				blit(page,screen,0,0,0,0,640,480);
			}
			if (keyboard_needs_poll()) poll_keyboard();
		}
	} while (false);
	textprintf_ex(screen, font, 0, 9, makecol(255,255,255), -1, "Game Over!");
	while (!keypressed()) {}
	destroy_bitmap(pieces);
	destroy_bitmap(hud);
	return 0;
}

END_OF_MAIN()
